import re
from typing import Union

import jdatetime
from django import forms
from django.core import exceptions, validators
from django.forms.utils import from_current_timezone, to_current_timezone
from django.utils.translation import gettext as _

from django_jalali.forms.widgets import jDateInput, jDateTimeInput


class jDateField(forms.Field):
    widget = jDateInput
    default_error_messages = {
        "invalid": _("Enter a valid date."),
    }

    def __init__(self, input_formats=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_formats = input_formats

    def to_python(self, value) -> Union[jdatetime.date, None]:
        """
        Validates that the input can be converted to a date. Returns a Python
        jdatetime.date object.
        """
        if value in validators.EMPTY_VALUES:
            return None
        if isinstance(value, jdatetime.datetime):
            return value.date()
        if isinstance(value, jdatetime.date):
            return value

        if self.input_formats:
            for input_format in self.input_formats:
                try:
                    return jdatetime.datetime.strptime(value, input_format).date()
                except ValueError:
                    pass

        groups = re.search(
            r"(?P<year>[\d]{1,4})-(?P<month>[\d]{1,2})-(?P<day>[\d]{1,2})",
            value,
        )
        try:
            return jdatetime.date(
                year=int(groups.group(1)),
                month=int(groups.group(2)),
                day=int(groups.group(3)),
            )

        except (ValueError, AttributeError):
            pass

        raise exceptions.ValidationError(self.error_messages["invalid"])


class jDateTimeField(forms.Field):
    widget = jDateTimeInput
    default_error_messages = {
        "invalid": _("Enter a valid date/time."),
    }

    def __init__(self, input_formats=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.input_formats = input_formats

    def prepare_value(self, value):
        if isinstance(value, jdatetime.datetime):
            value = to_current_timezone(value)
        return value

    def to_python(self, value):
        """
        Validates that the input can be converted to a datetime. Returns a
        Python datetime.datetime object.
        """
        if value in validators.EMPTY_VALUES:
            return None
        if isinstance(value, jdatetime.datetime):
            return from_current_timezone(value.togregorian())
        if isinstance(value, jdatetime.date):
            result = jdatetime.datetime(value.year, value.month, value.day)
            return from_current_timezone(result.togregorian())
        if isinstance(value, list):
            # Input comes from a SplitDateTimeWidget, for example. So, it's two
            # components: date and time.
            if len(value) != 2:
                raise exceptions.ValidationError(self.error_messages["invalid"])
            if (
                value[0] in validators.EMPTY_VALUES
                and value[1] in validators.EMPTY_VALUES
            ):
                return None
            value = "%s %s" % tuple(value)

        if self.input_formats:
            for input_format in self.input_formats:
                try:
                    return jdatetime.datetime.strptime(value, input_format).date()
                except ValueError:
                    pass

        groups = re.search(
            r"(?P<year>[\d]{1,4})-(?P<month>[\d]{1,2})-(?P<day>[\d]{1,2}) "
            r"(?P<hour>[\d]{1,2}):(?P<minute>[\d]{1,2})"
            r"(:(?P<second>[\d]{1,2}))?(.(?P<microsecond>[\d]{1,5}))?",
            value,
        )
        try:
            microsecond = int(groups.group("microsecond") or 0)
            second = int(groups.group("second") or 0)
            result = jdatetime.datetime(
                year=int(groups.group("year")),
                month=int(groups.group("month")),
                day=int(groups.group("day")),
                hour=int(groups.group("hour")),
                minute=int(groups.group("minute")),
                second=second,
                microsecond=microsecond,
            )
            return from_current_timezone(result.togregorian())

        except (ValueError, AttributeError):
            pass

        raise exceptions.ValidationError(self.error_messages["invalid"])



import django
import jdatetime
from django.contrib import admin
from django.contrib.admin.options import IncorrectLookupParameters
from django.core.exceptions import ValidationError
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from django_jalali.db import models


class JDateFieldListFilter(admin.FieldListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path):
        self.field_generic = "%s__" % field_path
        self.date_params = {
            k: v for k, v in params.items() if k.startswith(self.field_generic)
        }

        now = jdatetime.datetime.fromgregorian(datetime=timezone.now())
        if timezone.is_aware(now):
            now = timezone.localtime(now)

        if isinstance(field, models.jDateTimeField):
            today = now.replace(hour=0, minute=0, second=0, microsecond=0)
            format = "%Y-%m-%d %H:%M:%S"
        else:
            format = "%Y-%m-%d"
            today = now.date()

        tomorrow = today + jdatetime.timedelta(days=1)

        if today.month == 12:
            next_month = today.replace(year=today.year + 1, month=1, day=1)
        else:
            next_month = today.replace(month=today.month + 1, day=1)
        next_year = today.replace(year=today.year + 1, month=1, day=1)

        self.lookup_kwarg_since = "%s__gte" % field_path
        self.lookup_kwarg_until = "%s__lt" % field_path
        self.links = (
            (_("Any date"), {}),
            (
                _("Today"),
                {
                    self.lookup_kwarg_since: today.strftime(format),
                    self.lookup_kwarg_until: tomorrow.strftime(format),
                },
            ),
            (
                _("Past 7 days"),
                {
                    self.lookup_kwarg_since: (
                        today - jdatetime.timedelta(days=7)
                    ).strftime(format),
                    self.lookup_kwarg_until: tomorrow.strftime(format),
                },
            ),
            (
                _("This month"),
                {
                    self.lookup_kwarg_since: (today.replace(day=1)).strftime(format),
                    self.lookup_kwarg_until: next_month.strftime(format),
                },
            ),
            (
                _("This year"),
                {
                    self.lookup_kwarg_since: (today.replace(month=1, day=1)).strftime(
                        format
                    ),
                    self.lookup_kwarg_until: next_year.strftime(format),
                },
            ),
        )

        super().__init__(field, request, params, model, model_admin, field_path)

    def queryset(self, request, queryset):
        used_parameters = self.used_parameters.copy()
        if django.VERSION > (5, 0):
            used_parameters = {
                k: v[0] if isinstance(v, list) else v
                for k, v in used_parameters.items()
            }

        try:
            return queryset.filter(**used_parameters)
        except ValidationError as e:
            raise IncorrectLookupParameters(e)

    def expected_parameters(self):
        return [self.lookup_kwarg_since, self.lookup_kwarg_until]

    def choices(self, cl):
        date_params = self.date_params
        if django.VERSION > (5, 0):
            date_params = {
                k: v[0] if isinstance(v, list) else v for k, v in date_params.items()
            }

        for title, param_dict in self.links:
            yield {
                "selected": date_params == param_dict,
                "query_string": cl.get_query_string(param_dict, [self.field_generic]),
                "display": title,
            }

    def has_output(self):
        return True



from django.conf import settings
from django.core.signals import setting_changed

DEFAULTS = {
    # JavaScript static files for the admin Jalali date widget
    "ADMIN_JS_STATIC_FILES": [
        "admin/jquery.ui.datepicker.jalali/scripts/jquery-1.10.2.min.js",
        "admin/jquery.ui.datepicker.jalali/scripts/jquery.ui.core.js",
        "admin/jquery.ui.datepicker.jalali/scripts/jquery.ui.datepicker-cc.js",
        "admin/jquery.ui.datepicker.jalali/scripts/calendar.js",
        "admin/jquery.ui.datepicker.jalali/scripts/jquery.ui.datepicker-cc-fa.js",
        "admin/main.js",
    ],
    # CSS static files for the admin Jalali date widget
    "ADMIN_CSS_STATIC_FILES": {
        "all": [
            "admin/jquery.ui.datepicker.jalali/themes/base/jquery-ui.min.css",
            "admin/css/main.css",
        ]
    },
}


class JalaliSetting:
    def __init__(self, defaults=None):
        self._defaults = defaults or DEFAULTS
        self._attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "JALALI_SETTINGS", {})

        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self._defaults:
            raise AttributeError(f"Invalid Jalali setting: {attr}")

        try:
            val = self.user_settings[attr]
        except KeyError:
            val = self._defaults[attr]

        # Cache the result
        self._attrs.add(attr)
        setattr(self, attr, val)

        return val

    def reload(self):
        for attr in self._attrs:
            delattr(self, attr)

        self._attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


jalali_settings = JalaliSetting()


def reload_jalali_settings(*args, **kwargs):
    setting = kwargs["setting"]
    if setting == "JALALI_SETTINGS":
        jalali_settings.reload()


setting_changed.connect(reload_jalali_settings)



from django import forms
from django.contrib.admin.widgets import AdminTimeWidget
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

from django_jalali import forms as jforms
from django_jalali.settings import jalali_settings


class AdminjDateWidget(jforms.jDateInput):

    class Media:
        js = jalali_settings.ADMIN_JS_STATIC_FILES
        css = jalali_settings.ADMIN_CSS_STATIC_FILES

    def __init__(self, attrs=None, format=None):
        final_attrs = {"class": "vjDateField", "size": "10"}
        if attrs is not None:
            final_attrs.update(attrs)
        super().__init__(attrs=final_attrs, format=format)


class AdminSplitjDateTime(forms.SplitDateTimeWidget):
    """
    A SplitDateTime Widget that has some admin-specific styling.
    """

    def __init__(self, attrs=None):
        widgets = [AdminjDateWidget, AdminTimeWidget]
        # Note that we're calling MultiWidget, not SplitDateTimeWidget, because
        # we want to define widgets.
        forms.MultiWidget.__init__(self, widgets, attrs)

    def format_output(self, rendered_widgets):
        return mark_safe(
            '<p class="datetime">%s %s<br />%s %s</p>'
            % (
                _("Date:"),
                rendered_widgets[0],
                _("Time:"),
                rendered_widgets[1],
            )
        )



from django.test import TestCase, override_settings

from django_jalali.settings import DEFAULTS, jalali_settings


class TestJalaliSettings(TestCase):
    def test_default_settings_loaded(self):
        for key, value in DEFAULTS.items():
            self.assertEqual(getattr(jalali_settings, key), value)

    def test_override_settings(self):
        js = ["admin/scripts/custom.js"]
        with override_settings(JALALI_SETTINGS={"ADMIN_JS_STATIC_FILES": js}):
            self.assertEqual(jalali_settings.ADMIN_JS_STATIC_FILES, js)

        self.assertEqual(
            jalali_settings.ADMIN_JS_STATIC_FILES, DEFAULTS["ADMIN_JS_STATIC_FILES"]
        )

    def test_invalid_setting_access_raise_attribute_error(self):
        with self.assertRaises(AttributeError):
            jalali_settings.INVALID_SETTING



import jdatetime
from django.template import Context, Template
from django.test import TestCase


class JformatTestCase(TestCase):
    def setUp(self):
        date_time = jdatetime.date(1394, 11, 25)
        self.context = Context({"date_time": date_time})

    def test_jformat(self):
        _str = '{% load jformat %}{{ date_time|jformat:"%c" }}'
        output = "Sun Bah 25 00:00:00 1394"
        t = Template(_str)
        self.assertEqual(t.render(self.context), output)

    def test_jformat_unicode(self):
        _str = "{% load jformat %}"
        _str += '{{ date_time|jformat:"ﺱﺎﻟ = %y، ﻡﺎﻫ = %m، ﺭﻭﺯ = %d" }}'
        output = "ﺱﺎﻟ = 94، ﻡﺎﻫ = 11، ﺭﻭﺯ = 25"
        t = Template(_str)
        self.assertEqual(t.render(self.context), output)



import jdatetime
from django.test import RequestFactory, TestCase

from tests.models import Bar, BarTime
from tests.serializers import JDateFieldSerializer, JDateTimeFieldSerializer


class JDateFieldSerializerTests(TestCase):
    def setUp(self):
        self.request_factory = RequestFactory()
        self.today = jdatetime.date.today()

        # Bars
        self.mybar = Bar.objects.create(name="foo", date=self.today)

    def test_serializers_works_correctly_on_valid_date(self):
        serializer = JDateFieldSerializer(self.mybar)
        self.assertEqual(serializer.data["date"], str(self.today))

        serializer = JDateFieldSerializer(data={"name": "foo", "date": "1400-06-31"})
        self.assertTrue(serializer.is_valid())

    def test_serializers_works_correctly_on_leap_year(self):
        """
        Make sure JDateFieldSerializer work's correctly on leap and normal year
        """
        # Date is not a leap yaer, So this is not acceptable
        serializer = JDateFieldSerializer(
            data={"name": "leap-year", "date": "1400-12-30"}
        )
        self.assertFalse(serializer.is_valid())
        self.assertIn("date", serializer.errors)

        # Date is a leap year and is acceptable
        serializer = JDateFieldSerializer(
            data={"name": "leap-year", "date": "1399-12-30"}
        )
        self.assertTrue(serializer.is_valid())

    def test_serializer_save_date_correctly(self):
        data = {"name": "fooo", "date": "1400-12-29"}
        serializer = JDateFieldSerializer(data=data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        bar = Bar.objects.get(name="fooo")
        self.assertEqual(str(bar.date), data["date"])


class JDateTimeFieldSerializerTests(TestCase):
    def setUp(self):
        self.request_factory = RequestFactory()
        self.today = jdatetime.date.today()
        self.now = jdatetime.datetime.now()

        # Bars
        self.mybartime = BarTime.objects.create(name="bartime", datetime=self.now)

    def test_serializer_works_correctly_on_valid_datetime(self):
        serializer = JDateTimeFieldSerializer(self.mybartime)
        self.assertEqual(serializer.data["datetime"], str(self.now))

        serializer = JDateTimeFieldSerializer(
            data={"name": "bartime", "datetime": "1401-06-31"}
        )
        self.assertTrue(serializer.is_valid())

        serializer = JDateTimeFieldSerializer(
            data={"name": "bartime", "datetime": "1401-06-31 00:00"}
        )
        self.assertTrue(serializer.is_valid())

        serializer = JDateTimeFieldSerializer(
            data={"name": "bartime", "datetime": "1401-06-31 01:01:00"}
        )
        self.assertTrue(serializer.is_valid())

    def test_serializer_with_different_formatting(self):
        d_time = self.now.strftime("%Y-%m-%d %H:%M")
        bartime_object = BarTime.objects.create(name="formatted", datetime=d_time)
        self.assertEqual(
            JDateTimeFieldSerializer(bartime_object).data["datetime"], d_time
        )

    def test_serializer_with_invalid_datetime(self):
        serializer = JDateTimeFieldSerializer(
            data={"name": "bartime", "datetime": "1400-12-30 12:34:12"}
        )
        self.assertFalse(serializer.is_valid())

        serializer = JDateTimeFieldSerializer(
            data={"name": "bartime", "datetime": "1400-12-28 26:34:12"}
        )
        self.assertFalse(serializer.is_valid())



from urllib.parse import unquote

import jdatetime
from django.contrib.admin import site
from django.contrib.admin.views.main import ChangeList
from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory, TestCase
from django.utils.encoding import force_str

from tests.admin import BarTimeAdmin
from tests.models import BarTime


def select_by(dictlist, key, value):
    return [x for x in dictlist if x[key] == value][0]


class ListFiltersTests(TestCase):
    def setUp(self):
        self.request_factory = RequestFactory()
        self.today = jdatetime.date.today()
        self.tomorrow = self.today + jdatetime.timedelta(days=1)
        self.one_week_ago = self.today - jdatetime.timedelta(days=7)
        if self.today.month == 12:
            self.next_month = self.today.replace(
                year=self.today.year + 1, month=1, day=1
            )
        else:
            self.next_month = self.today.replace(month=self.today.month + 1, day=1)
        self.next_year = self.today.replace(year=self.today.year + 1, month=1, day=1)

        # Bars
        self.mybartime = BarTime.objects.create(name="foo", datetime=self.today)

    def test_jdatefieldlistfilter(self):
        modeladmin = BarTimeAdmin(BarTime, site)

        request = self.request_factory.get("/")
        request.user = AnonymousUser()
        changelist = self.get_changelist(request, BarTime, modeladmin)
        request = self.request_factory.get(
            "/",
            {
                "datetime__gte": self.today.strftime("%Y-%m-%d %H:%M:%S"),
                "datetime__lt": self.tomorrow.strftime("%Y-%m-%d %H:%M:%S"),
            },
        )
        request.user = AnonymousUser()
        changelist = self.get_changelist(request, BarTime, modeladmin)

        # Make sure the correct queryset is returned
        queryset = changelist.get_queryset(request)
        self.assertEqual(list(queryset), [self.mybartime])

        # Make sure the correct choice is selected
        filterspec = changelist.get_filters(request)[0][0]
        self.assertEqual(force_str(filterspec.title), "datetime")
        choice = select_by(filterspec.choices(changelist), "display", "Today")
        self.assertEqual(choice["selected"], True)

        self.assertEqual(
            unquote(choice["query_string"]).replace("+", " "),
            "?datetime__gte=%s&datetime__lt=%s"
            % (
                self.today.strftime("%Y-%m-%d %H:%M:%S"),
                self.tomorrow.strftime("%Y-%m-%d %H:%M:%S"),
            ),
        )

        request = self.request_factory.get(
            "/",
            {
                "datetime__gte": self.today.replace(day=1),
                "datetime__lt": self.next_month,
            },
        )
        request.user = AnonymousUser()
        changelist = self.get_changelist(request, BarTime, modeladmin)

    def get_changelist(self, request, model, modeladmin):
        args = [
            request,
            model,
            modeladmin.list_display,
            modeladmin.list_display_links,
            modeladmin.list_filter,
            modeladmin.date_hierarchy,
            modeladmin.search_fields,
            modeladmin.list_select_related,
            modeladmin.list_per_page,
            modeladmin.list_max_show_all,
            modeladmin.list_editable,
            modeladmin,
            modeladmin.sortable_by,
            modeladmin.search_help_text,
        ]
        return ChangeList(*args)



import jdatetime
from django.test import TestCase

from django_jalali.forms import jDateField, jDateTimeField


class JDateFieldTest(TestCase):
    def test_field(self):
        tests = (
            "1400-11-27",
            jdatetime.date(1400, 11, 27),
            jdatetime.datetime(1400, 11, 27),
        )
        for value in tests:
            with self.subTest(value=value):
                f = jDateField()
                self.assertEqual(f.clean(value), jdatetime.date(1400, 11, 27))

    def test_field_with_one_input_format(self):
        tests = (
            "1400/11/27",
            jdatetime.date(1400, 11, 27),
            jdatetime.datetime(1400, 11, 27),
        )
        for value in tests:
            with self.subTest(value=value):
                f = jDateField(
                    input_formats=[
                        "%Y/%m/%d",
                    ]
                )
                self.assertEqual(f.clean(value), jdatetime.date(1400, 11, 27))

    def test_field_with_multiple_input_formats(self):
        tests = (
            "1400/11/27",
            "1400, 11, 27",
            jdatetime.date(1400, 11, 27),
            jdatetime.datetime(1400, 11, 27),
        )
        for value in tests:
            with self.subTest(value=value):
                f = jDateField(input_formats=["%Y, %m, %d", "%Y/%m/%d"])
                self.assertEqual(f.clean(value), jdatetime.date(1400, 11, 27))


class JDateTimeFieldTest(TestCase):
    def test_field(self):
        tests = (
            "1400-11-27 12:13:20",
            jdatetime.datetime(1400, 11, 27, 12, 13, 20),
        )
        for value in tests:
            with self.subTest(value=value):
                f = jDateTimeField()
                self.assertEqual(
                    f.clean(value), jdatetime.datetime(1400, 11, 27, 12, 13, 20)
                )

    def test_field_with_one_input_formats(self):
        tests = (
            "1400/11/27 12:13",
            jdatetime.datetime(1400, 11, 27, 12, 13, 20),
        )
        for value in tests:
            with self.subTest(value=value):
                f = jDateTimeField(
                    input_formats=[
                        "%Y/%m/%d %H:%M",
                    ]
                )
                self.assertEqual(
                    f.clean(value), jdatetime.datetime(1400, 11, 27, 12, 13, 20)
                )

    def test_field_with_multiple_input_formats(self):
        tests = (
            "1400/11/27 12-13",
            "1400, 11, 27 12:13",
            jdatetime.datetime(1400, 11, 27),
        )
        for value in tests:
            with self.subTest(value=value):
                f = jDateTimeField(input_formats=["%Y, %m, %d %H:%M", "%Y/%m/%d %H-%M"])
                self.assertEqual(f.clean(value), jdatetime.date(1400, 11, 27))



import datetime

import jdatetime
from django.core.exceptions import ValidationError
from django.db.migrations.writer import MigrationWriter
from django.test import TestCase, override_settings, skipUnlessDBFeature
from django.test.utils import requires_tz_support
from django.utils import timezone

from django_jalali.db import models as jmodels
from tests.models import (
    Bar,
    BarTime,
    DateTimeWithDefault,
    DateWithDefault,
    ModelWithAutoNowAdd,
)


class BarTestCase(TestCase):
    def setUp(self):
        self.today_string = "1390-5-12"
        self.today = jdatetime.date(1390, 5, 12)
        self.mybar = Bar(name="foo", date=self.today)
        self.mybar.save()

    def test_save_date(self):
        self.assertEqual(self.mybar.date, self.today)

    def test_default_value(self):
        obj = DateWithDefault.objects.create()
        self.assertEqual(obj.date1, jdatetime.datetime(1390, 6, 31))
        self.assertEqual(obj.date2, jdatetime.datetime(1390, 6, 31))

    def test_save_specific_date(self):
        Bar.objects.create(name="Test", date="1398-04-31")
        k = Bar.objects.filter(date="1398-04-31")
        self.assertEqual(k[0].date.day, 31)

    def test_filter_by_exact_date(self):
        bars = Bar.objects.filter(date=self.today_string)
        self.assertEqual(len(bars), 1)

    def test_filter_by_exact_date_no_match(self):
        bars = Bar.objects.filter(date="1390-5-11")
        self.assertEqual(len(bars), 0)

    def test_filter_by_gte_date(self):
        bars = Bar.objects.filter(date__gte=self.today_string)
        self.assertEqual(len(bars), 1)

    def test_serialize_default_jdatetime_value(self):
        jd1 = jdatetime.date(1390, 6, 31)
        field = jmodels.jDateField(default=jd1)
        self.assertEqual(
            MigrationWriter.serialize(field),
            (
                "django_jalali.db.models.jDateField(default=datetime.date(2011, 9, 22))",
                {"import django_jalali.db.models", "import datetime"},
            ),
        )

    def test_serialize_default_datetime_value(self):
        dt1 = datetime.date(2013, 6, 2)
        field = jmodels.jDateField(default=dt1)
        self.assertEqual(
            MigrationWriter.serialize(field),
            (
                "django_jalali.db.models.jDateField(default=datetime.date(2013, 6, 2))",
                {"import django_jalali.db.models", "import datetime"},
            ),
        )

    def test_filtering(self):
        today = jdatetime.date.today()
        Bar.objects.create(date=today, name="test name")

        # search by year
        jdm_date = Bar.objects.filter(date__year=today.year)
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date.year, today.year)

        # search by string
        jdm_date = Bar.objects.filter(date=f"{today.year}-{today.month}-{today.day}")
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date, today)

        # save by geregorian date and retrive by jalali date
        g_2009 = datetime.date(2009, 9, 9)
        jd_model = Bar(date=g_2009)
        jd_model.save()

        jdm_date = Bar.objects.filter(date=g_2009)
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date, jdatetime.date.fromgregorian(date=g_2009))

        jdm_date = Bar.objects.filter(date=f"{today.year}-{today.month}-{today.day}")
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date, today)

        jdm_date = Bar.objects.filter(
            date__in=[f"{today.year}-{today.month}-{today.day}"]
        )
        self.assertEqual(len(jdm_date), 1)
        self.assertEqual(jdm_date[0].date, today)

        # wrong day value
        with self.assertRaises(ValidationError):
            Bar.objects.filter(date=f"{today.year}-{today.month}-{35}")

        # invalid format
        with self.assertRaises(ValidationError):
            Bar.objects.filter(date=f"{today.year}{today.month}/{35}")

        # invalid search
        with self.assertRaises(ValueError):
            Bar.objects.filter(date__month="test")


class BarTimeTestCase(TestCase):
    def setUp(self):
        self.date_string = "1380-08-02"
        self.datetime = jdatetime.datetime(1380, 8, 2, 12, 12, 12)
        self.bar_time = BarTime.objects.create(name="bar time", datetime=self.datetime)

    def test_save_datetime(self):
        self.assertEqual(self.bar_time.datetime, self.datetime)

    def test_default_value(self):
        obj = DateTimeWithDefault.objects.create()
        self.assertEqual(
            obj.datetime1, jdatetime.datetime(1390, 6, 31, 10, 22, 23, 240000)
        )
        self.assertEqual(
            obj.datetime2, jdatetime.datetime(1390, 6, 31, 10, 22, 23, 240000)
        )

    def test_save_specific_datetime(self):
        BarTime.objects.create(name="Test", datetime="1398-04-31 12:12:12")
        k = BarTime.objects.filter(datetime="1398-04-31 12:12:12")
        self.assertEqual(k[0].datetime.day, 31)

    def test_date_lookup_filter(self):
        jdatetime_ = jdatetime.datetime(1378, 10, 10)
        bartime = BarTime.objects.create(name="Test", datetime=jdatetime_)
        self.assertEqual(BarTime.objects.get(datetime__date=jdatetime_), bartime)
        self.assertEqual(BarTime.objects.get(datetime__date=jdatetime_.date()), bartime)
        self.assertEqual(BarTime.objects.get(datetime__date="1378-10-10"), bartime)
        with self.assertRaisesMessage(
            TypeError,
            "`__date` filter Expected jdatetime.datetime, jdatetime.date or str, got `int`.",
        ):
            BarTime.objects.get(datetime__date=1)

    def test_filter_by_exact_datetime(self):
        bar_times = BarTime.objects.filter(datetime=self.datetime)
        self.assertEqual(len(bar_times), 1)

    @requires_tz_support
    @skipUnlessDBFeature("has_zoneinfo_database")
    @override_settings(USE_TZ=True, TIME_ZONE="Asia/Tehran")
    def test_lookup_date_with_use_tz(self):
        current_timezone = timezone.get_current_timezone()
        jdt1 = jdatetime.datetime(
            1392, 3, 12, 10, 22, 23, 240000, tzinfo=current_timezone
        )

        BarTime.objects.create(name="with timezone", datetime=jdt1)
        k = BarTime.objects.filter(datetime=jdt1)
        self.assertEqual(str(k[0].datetime), "1392-03-12 10:22:23.240000+0430")
        self.assertEqual(k[0].datetime.strftime("%z"), "+0430")

    @requires_tz_support
    @override_settings(USE_TZ=True, TIME_ZONE="Asia/Tehran")
    def test_lookup_date_with_use_tz_without_explicit_tzinfo(self):
        jdt1 = jdatetime.datetime(1392, 3, 12, 10, 22, 23, 240000)
        BarTime.objects.create(name="with timezone", datetime=jdt1)
        k = BarTime.objects.filter(datetime=jdt1)
        self.assertEqual(str(k[0].datetime), "1392-03-12 10:22:23.240000+0430")
        self.assertEqual(k[0].datetime.strftime("%z"), "+0430")

    @requires_tz_support
    @override_settings(USE_TZ=False, TIME_ZONE="Asia/Tehran")
    def test_lookup_date_with_no_tz(self):
        jdt1 = jdatetime.datetime(1392, 3, 12, 10, 22, 23, 240000)
        BarTime.objects.create(name="with timezone", datetime=jdt1)
        k = BarTime.objects.filter(datetime=jdt1)
        self.assertEqual(str(k[0].datetime), "1392-03-12 10:22:23.240000")
        self.assertEqual(k[0].datetime.strftime("%z"), "")

    def test_lookup_auto_now_add(self):
        ModelWithAutoNowAdd.objects.create()
        dt = ModelWithAutoNowAdd.objects.all()[0].datetimefield
        objects = ModelWithAutoNowAdd.objects.filter(datetimefield=dt)
        self.assertEqual(objects[0].datetimefield, dt)

    @requires_tz_support
    @override_settings(USE_TZ=True, TIME_ZONE="Asia/Tehran")
    def test_lookup_auto_now_add_datetime_with_tz(self):
        ModelWithAutoNowAdd.objects.create()
        dt = ModelWithAutoNowAdd.objects.all()[0].datetimefield
        objects = ModelWithAutoNowAdd.objects.filter(datetimefield=dt)
        self.assertEqual(objects[0].datetimefield, dt)

    def test_serialize_default_jdatetime_value(self):
        jdt1 = jdatetime.datetime(1390, 6, 31, 10, 22, 23, 240000)
        field = jmodels.jDateTimeField(default=jdt1)
        self.assertEqual(
            MigrationWriter.serialize(field),
            (
                "django_jalali.db.models.jDateTimeField("
                "default=datetime.datetime(2011, 9, 22, 10, 22, 23, 240000))",
                {"import django_jalali.db.models", "import datetime"},
            ),
        )

    def test_serialize_default_datetime_value(self):
        dt1 = datetime.datetime(2013, 6, 2, 10, 22, 23, 240000)
        field = jmodels.jDateTimeField(default=dt1)
        self.assertEqual(
            MigrationWriter.serialize(field),
            (
                "django_jalali.db.models.jDateTimeField("
                "default=datetime.datetime(2013, 6, 2, 10, 22, 23, 240000))",
                {"import django_jalali.db.models", "import datetime"},
            ),
        )

    @requires_tz_support
    @override_settings(USE_TZ=True, TIME_ZONE="Asia/Tehran")
    def test_timezone(self):
        current_timezone = timezone.get_current_timezone()
        jdt1 = jdatetime.datetime(
            1392, 3, 12, 10, 22, 23, 240000, tzinfo=current_timezone
        )

        new_bartime = BarTime.objects.create(name="with timezone", datetime=jdt1)
        self.assertEqual(
            new_bartime.datetime.utcoffset(),
            datetime.timedelta(seconds=16200),
        )

        k = BarTime.objects.filter(datetime=jdt1)
        self.assertEqual(
            k[0].datetime.utcoffset(),
            datetime.timedelta(seconds=16200),
        )

    def test_chain_filters(self):
        qs = BarTime.objects.filter(name=self.bar_time.name).filter(
            datetime__year=self.datetime.year
        )
        self.assertEqual(qs.count(), 1)
